"""This script tests the run cli command"""
import logging
from pathlib import Path

from click.testing import CliRunner

from cg.cli.workflow.balsamic.base import run
from cg.constants import EXIT_SUCCESS
from cg.constants.priority import SlurmQos
from cg.models.cg_config import CGConfig


def test_without_options(cli_runner: CliRunner, balsamic_context: CGConfig):
    """Test command without case_id argument"""
    # GIVEN no case_id
    # WHEN dry running without anything specified
    result = cli_runner.invoke(run, obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN command should mention argument
    assert "Missing argument" in result.output


def test_with_missing_case(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with invalid case to start with"""
    caplog.set_level(logging.ERROR)
    # GIVEN case_id not in database
    case_id = "soberelephant"
    assert not balsamic_context.status_db.get_case_by_internal_id(internal_id=case_id)
    # WHEN running
    result = cli_runner.invoke(run, [case_id], obj=balsamic_context)
    # THEN command should NOT successfully call the command it creates
    assert result.exit_code != EXIT_SUCCESS
    # THEN ERROR log should be printed containing invalid case_id
    assert case_id in caplog.text
    assert "could not be found" in caplog.text


def test_without_samples(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with case_id and no samples"""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id
    case_id = "no_sample_case"
    # WHEN dry running with dry specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN warning should be printed that no config file is found
    assert case_id in caplog.text
    assert "no samples" in caplog.text


def test_without_config(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with case_id and no config file"""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id
    case_id = "balsamic_case_wgs_single"
    # WHEN dry running with dry specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS
    # THEN warning should be printed that no config file is found
    assert "No config file found" in caplog.text


def test_with_config(tmpdir_factory, cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with case_id and config file"""
    caplog.set_level(logging.INFO)
    # GIVEN case-id
    case_id = "balsamic_case_wgs_single"
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).parent,
        exist_ok=True,
    )
    Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).touch(
        exist_ok=True
    )
    # WHEN dry running with dry specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should NOT execute successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN warning should be printed that no config file is found
    assert "balsamic" in caplog.text


def test_run_analysis(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with dry run option"""
    caplog.set_level(logging.INFO)
    # GIVEN case-id
    case_id = "balsamic_case_wgs_single"
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).parent,
        exist_ok=True,
    )
    Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).touch(
        exist_ok=True
    )
    # WHEN dry running
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)
    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS


def test_priority_custom(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with priority option"""
    caplog.set_level(logging.INFO)
    # GIVEN valid case-id
    case_id = "balsamic_case_wgs_single"
    qos_key = "--slurm-quality-of-service"
    qos_value = SlurmQos.HIGH
    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).parent,
        exist_ok=True,
    )
    Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).touch(
        exist_ok=True
    )
    # WHEN dry running with option specified
    result = cli_runner.invoke(
        run, [case_id, "--dry-run", qos_key, qos_value], obj=balsamic_context
    )
    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS
    # THEN dry run should include the the priority value
    assert "--qos" in caplog.text
    assert qos_value in caplog.text


def test_priority_clinical(cli_runner: CliRunner, balsamic_context: CGConfig, caplog):
    """Test command with case_id set to default NORMAL priority, when priority is not set manually"""
    caplog.set_level(logging.INFO)

    # GIVEN valid case-id
    case_id = "balsamic_case_wgs_single"
    priority_value = SlurmQos.NORMAL

    # WHEN ensuring case config exists where it should be stored
    Path.mkdir(
        Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).parent,
        exist_ok=True,
    )
    Path(balsamic_context.meta_apis["analysis_api"].get_case_config_path(case_id)).touch(
        exist_ok=True
    )

    # WHEN dry running with option specified
    result = cli_runner.invoke(run, [case_id, "--dry-run"], obj=balsamic_context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN dry run should include the the priority value
    assert "--qos" in caplog.text
    assert priority_value in caplog.text



from typing import Any

from pydantic.v1 import BaseModel, conlist, constr

from cg.models.orders.constants import OrderType
from cg.models.orders.samples import sample_class_for
from cg.store.models import Customer, Sample


class OrderIn(BaseModel):
    name: constr(min_length=2, max_length=Sample.order.property.columns[0].type.length)
    comment: str | None
    customer: constr(min_length=1, max_length=Customer.internal_id.property.columns[0].type.length)
    samples: conlist(Any, min_items=1)
    skip_reception_control: bool | None = None
    ticket: str | None

    @classmethod
    def parse_obj(cls, obj: dict, project: OrderType) -> "OrderIn":
        parsed_obj: OrderIn = super().parse_obj(obj)
        parsed_obj.parse_samples(project=project)
        return parsed_obj

    def parse_samples(self, project: OrderType) -> None:
        """
        Parses samples of by the type given by the project

        Parameters:
            project (OrderType): type of project

        Returns:
            Nothing
        """
        parsed_samples = []

        sample: dict
        for sample in self.samples:
            parsed_sample = sample_class_for(project=project).parse_obj(sample)
            parsed_sample.skip_reception_control = self.skip_reception_control
            parsed_samples.append(parsed_sample)
        self.samples = parsed_samples



from pydantic import BaseModel

from cg.constants.constants import MicrosaltAppTags


class SampleQualityResult(BaseModel):
    sample_id: str
    passes_qc: bool
    is_control: bool
    application_tag: MicrosaltAppTags
    passes_reads_qc: bool
    passes_mapping_qc: bool = True
    passes_duplication_qc: bool = True
    passes_inserts_qc: bool = True
    passes_coverage_qc: bool = True
    passes_10x_coverage_qc: bool = True


class CaseQualityResult(BaseModel):
    passes_qc: bool
    control_passes_qc: bool
    urgent_passes_qc: bool
    non_urgent_passes_qc: bool


class QualityResult(BaseModel):
    case: CaseQualityResult
    samples: list[SampleQualityResult]
    summary: str

    @property
    def passes_qc(self) -> bool:
        return self.case.passes_qc



"""Symbol constants."""

ASTERISK = "*"
COMMA = ","
DASH = "-"
DOT = "."
EMPTY_STRING = ""
FWD_SLASH = "/"
LOWER_CASE_NA = "na"
NEW_LINE = "\n"
PERIOD = "."
SINGLE_QUOTE = "'"
SPACE = " "
UNDERSCORE = "_"



"""Tests for the CleanFlowCellsAPI."""
from datetime import datetime
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.subject import Sex
from cg.meta.clean.clean_flow_cells import CleanFlowCellAPI
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.store import Store
from cg.store.models import Flowcell, Sample
from tests.store_helpers import StoreHelpers


@pytest.fixture(scope="function")
def flow_cell_clean_api_can_be_removed(
    tmp_flow_cell_to_clean_path: Path,
    store_with_flow_cell_to_clean: Store,
    housekeeper_api_with_flow_cell_to_clean: HousekeeperAPI,
    tmp_sample_sheet_clean_flow_cell_path: Path,
) -> CleanFlowCellAPI:
    """Return a CleanFlowCellAPI with a flow cell that can be removed."""
    clean_flow_cell_api = CleanFlowCellAPI(
        flow_cell_path=tmp_flow_cell_to_clean_path,
        status_db=store_with_flow_cell_to_clean,
        housekeeper_api=housekeeper_api_with_flow_cell_to_clean,
        dry_run=False,
    )
    clean_flow_cell_api.flow_cell._sample_sheet_path_hk = tmp_sample_sheet_clean_flow_cell_path
    return clean_flow_cell_api


@pytest.fixture(scope="function")
def flow_cell_clean_api_can_not_be_removed(
    tmp_flow_cell_not_to_clean_path: Path,
    store_with_flow_cell_not_to_clean: Store,
    real_housekeeper_api: HousekeeperAPI,
) -> CleanFlowCellAPI:
    """Return a CleanFlowCellAPI with a flow cell that can not be removed."""
    return CleanFlowCellAPI(
        flow_cell_path=tmp_flow_cell_not_to_clean_path,
        status_db=store_with_flow_cell_not_to_clean,
        housekeeper_api=real_housekeeper_api,
        dry_run=False,
    )


@pytest.fixture(scope="function")
def tmp_flow_cell_to_clean_path(tmp_flow_cell_directory_bclconvert: Path):
    """Returns the path to a flow cell fulfilling all cleaning criteria."""
    return tmp_flow_cell_directory_bclconvert


@pytest.fixture(scope="function")
def tmp_flow_cell_to_clean(tmp_flow_cell_to_clean_path: Path) -> FlowCellDirectoryData:
    """Returns a flow cell directory object for a flow cell that fulfills all cleaning criteria."""
    return FlowCellDirectoryData(tmp_flow_cell_to_clean_path)


@pytest.fixture(scope="function")
def tmp_flow_cell_not_to_clean_path(tmp_flow_cell_directory_bcl2fastq: Path):
    """Return the path to a flow cell not fulfilling all cleaning criteria."""
    return tmp_flow_cell_directory_bcl2fastq


@pytest.fixture(scope="function")
def tmp_flow_cell_not_to_clean(tmp_flow_cell_not_to_clean_path: Path) -> FlowCellDirectoryData:
    """Returns a flow cell directory object for a flow cell that does not fulfill all cleaning criteria."""
    return FlowCellDirectoryData(tmp_flow_cell_not_to_clean_path)


@pytest.fixture(scope="session")
def tmp_sample_sheet_clean_flow_cell_path(tmp_path_factory) -> Path:
    sample_sheet_path = tmp_path_factory.mktemp("SampleSheet.csv")
    return sample_sheet_path


@pytest.fixture
def store_with_flow_cell_to_clean(
    store: Store,
    sample_id: str,
    tmp_flow_cell_to_clean: FlowCellDirectoryData,
    helpers: StoreHelpers,
) -> Store:
    """Return a store with multiple samples with sample lane sequencing metrics."""
    sample_sequencing_metrics_details: list[str | int | float] = [
        (sample_id, tmp_flow_cell_to_clean.id, 1, 50_000_0000, 90.5, 32),
        (sample_id, tmp_flow_cell_to_clean.id, 2, 50_000_0000, 90.4, 31),
    ]
    flow_cell: Flowcell = helpers.add_flow_cell(
        flow_cell_name=tmp_flow_cell_to_clean.id,
        store=store,
        has_backup=True,
    )
    sample: Sample = helpers.add_sample(
        store=store, customer_id="cust500", internal_id=sample_id, name=sample_id, sex=Sex.MALE
    )
    helpers.add_multiple_sample_lane_sequencing_metrics_entries(
        metrics_data=sample_sequencing_metrics_details, store=store
    )
    flow_cell.samples = [sample]
    store.session.add(flow_cell)
    store.session.commit()
    return store


@pytest.fixture
def store_with_flow_cell_not_to_clean(
    store: Store,
    sample_id: str,
    tmp_flow_cell_not_to_clean: FlowCellDirectoryData,
    helpers: StoreHelpers,
) -> Store:
    """Return a store with multiple samples with sample lane sequencing metrics."""
    sample_sequencing_metrics_details: list[str | int | float] = [
        (sample_id, tmp_flow_cell_not_to_clean.id, 1, 50_000_0000, 90.5, 32),
        (sample_id, tmp_flow_cell_not_to_clean.id, 2, 50_000_0000, 90.4, 31),
    ]
    flow_cell: Flowcell = helpers.add_flow_cell(
        flow_cell_name=tmp_flow_cell_not_to_clean.id,
        store=store,
        has_backup=True,
    )
    sample: Sample = helpers.add_sample(
        store=store, customer_id="cust500", internal_id=sample_id, name=sample_id, sex=Sex.MALE
    )
    helpers.add_multiple_sample_lane_sequencing_metrics_entries(
        metrics_data=sample_sequencing_metrics_details, store=store
    )
    flow_cell.samples = [sample]
    store.session.add(flow_cell)
    store.session.commit()
    return store


@pytest.fixture(scope="function")
def housekeeper_api_with_flow_cell_to_clean(
    real_housekeeper_api: HousekeeperAPI,
    helpers: StoreHelpers,
    hk_flow_cell_to_clean_bundle: dict,
    hk_sample_bundle_for_flow_cell_to_clean: dict,
) -> HousekeeperAPI:
    """
    Return a housekeeper api that contains a flow cell bundle with sample sheet,
    a sample bundle with a fastq and a SPRING file that are tagged with the flow cell.
    """
    helpers.ensure_hk_bundle(store=real_housekeeper_api, bundle_data=hk_flow_cell_to_clean_bundle)
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api, bundle_data=hk_sample_bundle_for_flow_cell_to_clean
    )
    return real_housekeeper_api


@pytest.fixture(scope="function")
def housekeeper_api_with_flow_cell_not_to_clean(
    real_housekeeper_api: HousekeeperAPI,
    helpers: StoreHelpers,
    hk_sample_bundle_for_flow_cell_not_to_clean: dict,
) -> HousekeeperAPI:
    """
    Return a housekeeper api that contains a flow cell bundle with sample sheet,
    a sample bundle with a fastq and a SPRING file that are tagged with the flow cell.
    """
    helpers.ensure_hk_bundle(
        store=real_housekeeper_api, bundle_data=hk_sample_bundle_for_flow_cell_not_to_clean
    )
    return real_housekeeper_api


@pytest.fixture(scope="function")
def hk_flow_cell_to_clean_bundle(
    tmp_flow_cell_to_clean: FlowCellDirectoryData,
    timestamp_yesterday: datetime,
    tmp_sample_sheet_clean_flow_cell_path: Path,
) -> dict:
    """Housekeeper bundle information for a flow cell that can be cleaned."""
    return {
        "name": tmp_flow_cell_to_clean.id,
        "created": timestamp_yesterday,
        "expires": timestamp_yesterday,
        "files": [
            {
                "path": str(tmp_sample_sheet_clean_flow_cell_path),
                "archive": False,
                "tags": ["samplesheet", tmp_flow_cell_to_clean.id],
            }
        ],
    }


@pytest.fixture(scope="function")
def hk_sample_bundle_for_flow_cell_to_clean(
    sample_id: str,
    timestamp_yesterday: datetime,
    spring_file: Path,
    fastq_file: Path,
    spring_meta_data_file: Path,
    tmp_flow_cell_to_clean: FlowCellDirectoryData,
) -> dict:
    return {
        "name": sample_id,
        "created": timestamp_yesterday,
        "expires": timestamp_yesterday,
        "files": [
            {
                "path": spring_file.as_posix(),
                "archive": False,
                "tags": [SequencingFileTag.SPRING, sample_id, tmp_flow_cell_to_clean.id],
            },
            {
                "path": fastq_file.as_posix(),
                "archive": False,
                "tags": [SequencingFileTag.FASTQ, sample_id, tmp_flow_cell_to_clean.id],
            },
            {
                "path": spring_meta_data_file.as_posix(),
                "archive": False,
                "tags": [SequencingFileTag.SPRING_METADATA, sample_id, tmp_flow_cell_to_clean.id],
            },
        ],
    }


@pytest.fixture(scope="function")
def hk_sample_bundle_for_flow_cell_not_to_clean(
    sample_id: str,
    timestamp_yesterday: datetime,
) -> dict:
    return {
        "name": sample_id,
        "created": timestamp_yesterday,
        "expires": timestamp_yesterday,
        "files": [],
    }



import logging

import click

from cg.meta.upload.fohm.fohm import FOHMUploadAPI
from cg.meta.upload.gisaid import GisaidAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)

ARGUMENT_DATE = click.argument(
    "datestr",
    type=str,
    required=False,
)
OPTION_CASES = click.option(
    "cases",
    "-c",
    multiple=True,
    type=str,
    required=True,
    help="CG internal id of cases to aggregate for daily delivery",
)
OPTION_DRY_RUN = click.option(
    "dry_run", "--dry-run", is_flag=True, default=False, help="Run command in dry run"
)


@click.group()
@click.pass_obj
def fohm(context: CGConfig):
    pass


@fohm.command("aggregate-delivery")
@OPTION_CASES
@OPTION_DRY_RUN
@ARGUMENT_DATE
@click.pass_obj
def aggregate_delivery(
    context: CGConfig, cases: list, dry_run: bool = False, datestr: str | None = None
):
    """Re-aggregates delivery files for FOHM and saves them to default working directory"""
    fohm_api = FOHMUploadAPI(config=context, dry_run=dry_run, datestr=datestr)
    fohm_api.set_cases_to_aggregate(cases=cases)
    fohm_api.create_daily_delivery_folders()
    fohm_api.append_metadata_to_aggregation_df()
    fohm_api.create_komplettering_reports()
    fohm_api.create_pangolin_reports()
    fohm_api.link_sample_rawdata_files()


@fohm.command("create-komplettering")
@OPTION_CASES
@OPTION_DRY_RUN
@ARGUMENT_DATE
@click.pass_obj
def create_komplettering(
    context: CGConfig, cases: list, dry_run: bool = False, datestr: str | None = None
):
    """Re-aggregates komplettering files for FOHM and saves them to default working directory"""
    fohm_api = FOHMUploadAPI(config=context, dry_run=dry_run, datestr=datestr)
    fohm_api.set_cases_to_aggregate(cases=cases)
    fohm_api.create_daily_delivery_folders()
    fohm_api.append_metadata_to_aggregation_df()
    fohm_api.create_komplettering_reports()


@fohm.command("preprocess-all")
@OPTION_CASES
@OPTION_DRY_RUN
@ARGUMENT_DATE
@click.pass_obj
def preprocess_all(
    context: CGConfig, cases: list, dry_run: bool = False, datestr: str | None = None
):
    """Create all FOHM upload files, upload to GISAID, sync SFTP and mail reports for all provided cases"""
    fohm_api = FOHMUploadAPI(config=context, dry_run=dry_run, datestr=datestr)
    gisaid_api = GisaidAPI(config=context)
    cases = list(cases)
    upload_cases = []
    for case_id in cases:
        try:
            gisaid_api.upload(case_id=case_id)
            fohm_api.update_upload_started_at(case_id=case_id)
            LOG.info(f"Upload of case {case_id} to GISAID was successful")
            upload_cases.append(case_id)
        except Exception as error:
            LOG.error(
                f"Upload of case {case_id} to GISAID unseccessful {error}, case {case_id} "
                f"will be removed from delivery batch"
            )
    fohm_api.set_cases_to_aggregate(cases=upload_cases)
    fohm_api.create_daily_delivery_folders()
    fohm_api.append_metadata_to_aggregation_df()
    fohm_api.create_komplettering_reports()
    fohm_api.create_pangolin_reports()
    fohm_api.link_sample_rawdata_files()
    fohm_api.sync_files_sftp()
    fohm_api.send_mail_reports()
    for case_id in upload_cases:
        fohm_api.update_uploaded_at(case_id=case_id)
    LOG.info("Upload to FOHM completed")


@fohm.command("upload-rawdata")
@ARGUMENT_DATE
@OPTION_DRY_RUN
@click.pass_obj
def upload_rawdata(context: CGConfig, dry_run: bool = False, datestr: str | None = None):
    """Deliver files in daily upload directory via sftp"""
    fohm_api = FOHMUploadAPI(config=context, dry_run=dry_run, datestr=datestr)
    fohm_api.sync_files_sftp()


@fohm.command("send-reports")
@ARGUMENT_DATE
@OPTION_DRY_RUN
@click.pass_obj
def send_reports(context: CGConfig, dry_run: bool = False, datestr: str | None = None):
    """Send all komplettering reports found in current daily directory to target recipients"""
    fohm_api = FOHMUploadAPI(config=context, dry_run=dry_run, datestr=datestr)
    fohm_api.send_mail_reports()



import logging
from pathlib import Path

from housekeeper.store.models import File
from pydantic import BaseModel, Field
from requests import Response

from cg.constants.constants import APIMethods
from cg.io.controller import APIRequest
from cg.meta.archive.ddn.constants import OSTYPE, ROOT_TO_TRIM, JobStatus
from cg.meta.archive.models import FileTransferData
from cg.store.models import Sample

LOG = logging.getLogger(__name__)


def get_request_log(headers: dict, body: dict):
    return "Sending request with headers: \n" + f"{headers} \n" + "and body: \n" + f"{body}"


class MiriaObject(FileTransferData):
    """Model for representing a singular object transfer."""

    destination: str
    source: str

    @classmethod
    def create_from_file_and_sample(
        cls, file: File, sample: Sample, is_archiving: bool = True
    ) -> "MiriaObject":
        """Instantiates the class from a File and Sample object."""
        if is_archiving:
            return cls(destination=sample.internal_id, source=file.full_path)
        return cls(
            destination=Path(file.full_path).parent.as_posix(),
            source=Path(sample.internal_id, Path(file.path).name).as_posix(),
        )

    def trim_path(self, attribute_to_trim: str):
        """Trims the given attribute (source or destination) from its root directory."""
        setattr(
            self,
            attribute_to_trim,
            f"/{Path(getattr(self, attribute_to_trim)).relative_to(ROOT_TO_TRIM)}",
        )

    def add_repositories(self, source_prefix: str, destination_prefix: str):
        """Prepends the given repositories to the source and destination paths."""
        self.source: str = source_prefix + self.source
        self.destination: str = destination_prefix + self.destination


class TransferPayload(BaseModel):
    """Model for representing a Dataflow transfer task."""

    files_to_transfer: list[MiriaObject]
    osType: str = OSTYPE
    createFolder: bool = True
    settings: list[dict] = []
    metadataList: list[dict] = []

    def trim_paths(self, attribute_to_trim: str):
        """Trims the source path from its root directory for all objects in the transfer."""
        for miria_file in self.files_to_transfer:
            miria_file.trim_path(attribute_to_trim=attribute_to_trim)

    def add_repositories(self, source_prefix: str, destination_prefix: str):
        """Prepends the given repositories to the source and destination paths all objects in the
        transfer."""
        for miria_file in self.files_to_transfer:
            miria_file.add_repositories(
                source_prefix=source_prefix, destination_prefix=destination_prefix
            )

    def model_dump(self, **kwargs) -> dict:
        """Creates a correctly structured dict to be used as the request payload."""
        payload: dict = super().model_dump(exclude={"files_to_transfer"})
        payload["pathInfo"] = [miria_file.model_dump() for miria_file in self.files_to_transfer]
        return payload

    def post_request(self, url: str, headers: dict) -> "TransferResponse":
        """Sends a request to the given url with, the given headers, and its own content as payload.

        Arguments:
            url: URL to which the POST goes to.
            headers: Headers which are set in the request
        Raises:
            HTTPError if the response status is not successful.
            ValidationError if the response does not conform to the expected response structure.
        Returns:
            The job ID of the launched transfer task.
        """

        LOG.info(get_request_log(headers=headers, body=self.model_dump()))

        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.POST,
            url=url,
            headers=headers,
            json=self.model_dump(),
            verify=False,
        )
        response.raise_for_status()
        return TransferResponse.model_validate(response.json())


class TransferResponse(BaseModel):
    """Model representing th response fields of an archive or retrieve reqeust to the Dataflow
    API."""

    job_id: int = Field(alias="jobId")


class AuthPayload(BaseModel):
    """Model representing the payload for an Authentication request."""

    dbName: str
    name: str
    password: str
    superUser: bool = False


class RefreshPayload(BaseModel):
    """Model representing the payload for Auth-token refresh request."""

    refresh: str


class AuthToken(BaseModel):
    """Model representing the response fields from an access request to the Dataflow API."""

    access: str
    expire: int
    refresh: str | None = None


class GetJobStatusResponse(BaseModel):
    """Model representing the response fields from a get_job_status post."""

    job_id: int = Field(alias="id")
    status: JobStatus


class GetJobStatusPayload(BaseModel):
    """Model representing the payload for a get_job_status request."""

    id: int

    def get_job_status(self, url: str, headers: dict) -> GetJobStatusResponse:
        """Sends a get request to the given URL with the given headers.
        Returns the parsed status response of the task specified in the URL.
        Raises:
             HTTPError if the response code is not ok.
        """

        LOG.info(get_request_log(headers=headers, body=self.model_dump()))

        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.GET,
            url=url,
            headers=headers,
            json=self.model_dump(),
            verify=False,
        )
        response.raise_for_status()
        return GetJobStatusResponse.model_validate(response.json())


class DeleteFileResponse(BaseModel):
    message: str


class DeleteFilePayload(BaseModel):
    global_path: str

    def delete_file(self, url: str, headers: dict) -> DeleteFileResponse:
        """Posts to the given URL with the given headers to delete the file or directory at the specified global path.
        Returns the parsed response.
        Raises:
             HTTPError if the response code is not ok.
        """
        LOG.info(get_request_log(headers=headers, body=self.model_dump()))

        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.POST,
            url=url,
            headers=headers,
            json=self.model_dump(),
            verify=False,
        )
        response.raise_for_status()
        return DeleteFileResponse.model_validate(response.json())



import logging
from pathlib import Path

from click.testing import CliRunner, Result

import cg
from cg.cli.base import base, init
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.database import get_tables, initialize_database


def test_cli_version(cli_runner: CliRunner):
    # GIVEN I want to see the version of the program
    # WHEN asking to see the version
    result: Result = cli_runner.invoke(base, ["--version"])
    # THEN it should display the version of the program
    # THEN it should print the name and version of the tool only
    assert cg.__title__ in result.output
    assert cg.__version__ in result.output


def test_list_commands(cli_runner: CliRunner):
    # WHEN using simplest command 'cg'
    result = cli_runner.invoke(base, [])
    # THEN it should just work ;-)
    assert result.exit_code == 0


def test_missing_command(cli_runner: CliRunner):
    # WHEN invoking a missing command
    result = cli_runner.invoke(base, ["i_dont_exist"])
    # THEN context should abort
    assert result.exit_code != 0


def test_cli_init(cli_runner: CliRunner, base_context: CGConfig, caplog, tmp_path: Path):
    caplog.set_level(logging.INFO)
    # GIVEN you want to setup a new database using the CLI
    database = Path(tmp_path, "test_db.sqlite3")
    database_path = Path(database)
    database_uri = f"sqlite:///{database}"
    initialize_database(database_uri)
    base_context.status_db_ = Store()

    assert database_path.exists() is False

    # WHEN calling "init"
    result = cli_runner.invoke(init, [], obj=base_context)

    # THEN it should setup the database with some tables
    assert result.exit_code == 0
    assert database_path.exists()
    assert len(get_tables()) > 0

    # GIVEN the database already exists
    # WHEN calling the init function
    result = cli_runner.invoke(init, [], obj=base_context)

    # THEN it should print an error and give error exit code
    assert result.exit_code != 0
    assert "Database already exists" in caplog.text

    # GIVEN the database already exists
    # WHEN calling "init" with "--reset"
    result = cli_runner.invoke(init, ["--reset"], input="Yes", obj=base_context)

    # THEN it should re-setup the tables and print new tables
    assert result.exit_code == 0
    assert "Success!" in caplog.text



from typing import Any, Callable


class Dispatcher:
    def __init__(self, functions: list[Callable], input_dict: dict[str, Any]):
        """Initialize the dispatcher with a list of functions and an input dict"""
        self.functions: list[Callable] = functions
        self.input_dict: dict[str, any] = input_dict
        self.dispatch_table: dict[tuple[str], Callable] = self._generate_dispatch_table()

    def __call__(self):
        """Call the dispatcher with the input dict"""
        parameters_not_none, parameter_values_not_none = self._parse_input_dict(self.input_dict)
        if parameters_not_none not in self.dispatch_table:
            raise ValueError(f"No matching function found for arguments: {parameters_not_none}")
        return self.dispatch_table[parameters_not_none](
            **dict(zip(parameters_not_none, parameter_values_not_none))
        )

    def _parse_input_dict(self, input_dict: dict[str, Any]) -> [tuple[str], tuple[Any]]:
        """Parse the input dict and return a tuple of parameters and values that are not None"""
        parameters_not_none: list[str] = []
        values_not_none: list[Any] = []
        for parameter in sorted(input_dict.keys()):
            value = input_dict[parameter]
            if value is not None:
                parameters_not_none.append(parameter)
                values_not_none.append(value)
        return tuple(parameters_not_none), tuple(values_not_none)

    def _generate_dispatch_table(self) -> dict[tuple[str], Callable]:
        """Generate a dispatch table from the list of functions"""
        dispatch_table = {}
        for func in self.functions:
            func_arg_names: list[str] = sorted(
                func.__code__.co_varnames[: func.__code__.co_argcount]
            )
            func_arg_names: list[str] = [arg for arg in func_arg_names if arg != "self"]
            if len(set(func_arg_names)) != len(func_arg_names):
                raise ValueError(f"Duplicate argument names in function: {func.__name__}")
            dispatch_table[tuple(func_arg_names)] = func
        return dispatch_table



"""Module to parse dates."""
import logging
import re
from datetime import datetime, timedelta

from cg.constants.symbols import DASH, DOT, FWD_SLASH, SPACE

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
DATETIME_FORMAT_DATE = "%Y-%m-%d %H:%M:%S"
SIMPLE_DATE_FORMAT = "%Y-%m-%d"

LOG = logging.getLogger(__name__)


def match_date(date: str) -> bool:
    """Check if a string is a valid date."""
    date_pattern = re.compile(r"^(19|20)\d\d[- /.](0[1-9]|1[012])[- /.](0[1-9]|[12][0-9]|3[01])")
    return bool(re.match(date_pattern, date))


def get_date(date: str | None = None, date_format: str | None = None) -> datetime:
    """Return a datetime object if there is a valid date.

    Raise exception if date is not valid.
    Return todays date if no date where added.

    """
    LOG.info(f"Trying to parse date string {date}")
    if not date:
        return datetime.now()

    if date_format:
        return datetime.strptime(date, date_format)

    if not match_date(date):
        raise ValueError(f"Date {date} is invalid")

    try:
        return datetime.strptime(date, DATETIME_FORMAT)
    except ValueError:
        try:
            return datetime.strptime(date, DATETIME_FORMAT_DATE)
        except ValueError:
            LOG.info("Date is not in standard format")

    for separator in [DASH, SPACE, DOT, FWD_SLASH]:
        date_parts = date.split(separator)
        if len(date_parts) == 3:
            return datetime(*(int(number) for number in date_parts))
    raise ValueError(f"Date {date} is invalid")


def get_timedelta_from_date(date: datetime) -> timedelta:
    """Return the number of days ago from date to now."""
    return datetime.now() - date


def get_date_days_ago(days_ago: int) -> datetime:
    """Return the date that was number of 'days_ago'."""
    return datetime.now() - timedelta(days=days_ago)



"""Tests for the SampleSheetCreator classes."""
from pathlib import Path

import pytest

from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_validated_sample_sheet
from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreator,
    SampleSheetCreatorBcl2Fastq,
    SampleSheetCreatorBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.constants.demultiplexing import BclConverter
from cg.exc import SampleSheetError
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData


def test_bcl_convert_sample_sheet_fails_with_bcl2fastq(
    novaseq_x_flow_cell: FlowCellDirectoryData,
    lims_novaseq_bcl_convert_samples: list[FlowCellSampleBCLConvert],
):
    """Test that creating a BCL Convert sample sheet fails if the bcl converter is Bcl2fastq."""
    # GIVEN a NovaSeqX flow cell and samples and the bcl converter is Bcl2fastq
    novaseq_x_flow_cell.bcl_converter = BclConverter.BCL2FASTQ

    # WHEN trying to instantiate a SampleSheetCreatorBCLConvert with Bcl2fastq as bcl_converter
    with pytest.raises(SampleSheetError) as exc_info:
        SampleSheetCreatorBCLConvert(
            flow_cell=novaseq_x_flow_cell,
            lims_samples=lims_novaseq_bcl_convert_samples,
        )
        # THEN an error is raised
        assert (
            str(exc_info.value)
            == f"Can't use {BclConverter.BCL2FASTQ} with BCL Convert sample sheet"
        )


def test_construct_bcl2fastq_sheet(
    bcl2fastq_sample_sheet_creator: SampleSheetCreatorBcl2Fastq, project_dir: Path
):
    """Test that a created Bcl2fastq sample sheet has samples."""
    # GIVEN a Bcl2fastq sample sheet creator populated with Bcl2fastq samples
    assert bcl2fastq_sample_sheet_creator.lims_samples

    # WHEN building the sample sheet
    sample_sheet_content: list[list[str]] = bcl2fastq_sample_sheet_creator.construct_sample_sheet()

    # THEN a correctly formatted sample sheet was created
    sample_sheet: SampleSheet = get_validated_sample_sheet(
        sample_sheet_content=sample_sheet_content,
        sample_type=FlowCellSampleBcl2Fastq,
    )
    assert sample_sheet.samples


def test_construct_bcl_convert_sheet(
    bcl_convert_sample_sheet_creator: SampleSheetCreator, project_dir: Path
):
    """Test that a created BCL Convert sample sheet has samples."""
    # GIVEN a BCL convert sample sheet creator populated with BCL convert samples
    assert bcl_convert_sample_sheet_creator.lims_samples

    # WHEN building the sample sheet
    sample_sheet_content: list[
        list[str]
    ] = bcl_convert_sample_sheet_creator.construct_sample_sheet()

    # THEN a correctly formatted sample sheet was created
    sample_sheet: SampleSheet = get_validated_sample_sheet(
        sample_sheet_content=sample_sheet_content,
        sample_type=FlowCellSampleBCLConvert,
    )
    assert sample_sheet.samples


def test_remove_unwanted_samples_dual_index(
    novaseq6000_flow_cell_sample_before_adapt_indexes: FlowCellSampleBcl2Fastq,
    hiseq_x_single_index_flow_cell: FlowCellDirectoryData,
):
    """Test that a sample with dual index is not removed."""
    # GIVEN a sample sheet creator with a sample with dual index
    sample_sheet_creator: SampleSheetCreatorBcl2Fastq = SampleSheetCreatorBcl2Fastq(
        flow_cell=hiseq_x_single_index_flow_cell,
        lims_samples=[novaseq6000_flow_cell_sample_before_adapt_indexes],
    )

    # WHEN removing unwanted samples
    sample_sheet_creator.remove_samples_with_simple_index()

    # THEN the sample is not removed
    assert len(sample_sheet_creator.lims_samples) == 1


def test_remove_unwanted_samples_no_dual_index(
    novaseq6000_flow_cell_sample_no_dual_index: FlowCellSampleBcl2Fastq,
    novaseq_6000_flow_cell: FlowCellDirectoryData,
    caplog,
):
    """Test that samples with no dual index are removed."""
    # GIVEN a sample sheet creator with a sample without dual indexes
    sample_sheet_creator: SampleSheetCreatorBcl2Fastq = SampleSheetCreatorBcl2Fastq(
        flow_cell=novaseq_6000_flow_cell,
        lims_samples=[novaseq6000_flow_cell_sample_no_dual_index],
    )

    # WHEN removing unwanted samples
    sample_sheet_creator.remove_samples_with_simple_index()

    # THEN the only sample is removed
    assert len(sample_sheet_creator.lims_samples) == 0
    assert (
        f"Removing sample {novaseq6000_flow_cell_sample_no_dual_index.sample_id} since it does not have dual index"
        in caplog.text
    )



"""Module to create delivery reports."""
import logging
from datetime import datetime
from pathlib import Path

import requests
from housekeeper.store.models import File, Version
from jinja2 import Environment, PackageLoader, Template, select_autoescape
from sqlalchemy.orm import Query

from cg.constants import Pipeline
from cg.constants.constants import MAX_ITEMS_TO_RETRIEVE, FileFormat
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG
from cg.exc import DeliveryReportError
from cg.io.controller import WriteStream
from cg.meta.meta import MetaAPI
from cg.meta.report.field_validators import (
    get_empty_report_data,
    get_missing_report_data,
)
from cg.meta.workflow.analysis import AnalysisAPI
from cg.models.analysis import AnalysisModel
from cg.models.cg_config import CGConfig
from cg.models.report.metadata import SampleMetadataModel
from cg.models.report.report import (
    CaseModel,
    CustomerModel,
    DataAnalysisModel,
    ReportModel,
    ScoutReportFiles,
)
from cg.models.report.sample import (
    ApplicationModel,
    MethodsModel,
    SampleModel,
    TimestampModel,
)
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    Case,
    CaseSample,
    Sample,
)

LOG = logging.getLogger(__name__)


class ReportAPI(MetaAPI):
    """Common Delivery Report API."""

    def __init__(self, config: CGConfig, analysis_api: AnalysisAPI):
        super().__init__(config=config)
        self.analysis_api: AnalysisAPI = analysis_api

    def create_delivery_report(
        self, case_id: str, analysis_date: datetime, force_report: bool
    ) -> str:
        """Generates the html contents of a delivery report."""
        report_data: ReportModel = self.get_report_data(
            case_id=case_id, analysis_date=analysis_date
        )
        report_data: ReportModel = self.validate_report_fields(
            case_id=case_id, report_data=report_data, force_report=force_report
        )
        rendered_report: str = self.render_delivery_report(report_data=report_data.model_dump())
        return rendered_report

    def create_delivery_report_file(
        self, case_id: str, directory: Path, analysis_date: datetime, force_report: bool
    ) -> Path:
        """Generates a file containing the delivery report content."""
        directory.mkdir(parents=True, exist_ok=True)
        delivery_report: str = self.create_delivery_report(
            case_id=case_id, analysis_date=analysis_date, force_report=force_report
        )
        report_file_path: Path = Path(directory, "delivery-report.html")
        with open(report_file_path, "w") as delivery_report_stream:
            delivery_report_stream.write(delivery_report)
        return report_file_path

    def add_delivery_report_to_hk(
        self, case_id: str, delivery_report_file: Path, version: Version
    ) -> File | None:
        """Add a delivery report file to a case bundle and return its file object."""
        LOG.info(f"Adding a new delivery report to housekeeper for {case_id}")
        file: File = self.housekeeper_api.add_file(
            path=delivery_report_file, version_obj=version, tags=[case_id, HK_DELIVERY_REPORT_TAG]
        )
        self.housekeeper_api.include_file(file, version)
        self.housekeeper_api.add_commit(file)
        return file

    def get_delivery_report_from_hk(self, case_id: str, version: Version) -> str | None:
        """Return path of a delivery report stored in HK."""
        delivery_report: File = self.housekeeper_api.get_latest_file(
            bundle=case_id, tags=[HK_DELIVERY_REPORT_TAG], version=version.id
        )
        if not delivery_report:
            LOG.warning(f"No delivery report found in housekeeper for {case_id}")
            return None
        return delivery_report.full_path

    def get_scout_uploaded_file_from_hk(self, case_id: str, scout_tag: str) -> str | None:
        """Return file path of the uploaded to Scout file given its tag."""
        version: Version = self.housekeeper_api.last_version(bundle=case_id)
        tags: list = self.get_hk_scout_file_tags(scout_tag=scout_tag)
        uploaded_file: File = self.housekeeper_api.get_latest_file(
            bundle=case_id, tags=tags, version=version.id
        )
        if not tags or not uploaded_file:
            LOG.warning(
                f"No files were found for the following Scout Housekeeper tag: {scout_tag} (case: {case_id})"
            )
            return None
        return uploaded_file.full_path

    def render_delivery_report(self, report_data: dict) -> str:
        """Renders the report on the Jinja template."""
        env: Environment = Environment(
            loader=PackageLoader("cg", "meta/report/templates"),
            autoescape=select_autoescape(["html", "xml"]),
        )
        template: Template = env.get_template(self.get_template_name())
        return template.render(**report_data)

    def get_cases_without_delivery_report(self, pipeline: Pipeline) -> list[Case]:
        """Returns a list of cases that has been stored and need a delivery report."""
        stored_cases: list[Case] = []
        analyses: Query = self.status_db.analyses_to_delivery_report(pipeline=pipeline)[
            :MAX_ITEMS_TO_RETRIEVE
        ]
        for analysis_obj in analyses:
            case: Case = analysis_obj.case
            last_version: Version = self.housekeeper_api.last_version(bundle=case.internal_id)
            hk_file: File = self.housekeeper_api.get_files(
                bundle=case.internal_id, version=last_version.id if last_version else None
            ).first()

            if hk_file and Path(hk_file.full_path).is_file():
                stored_cases.append(case)
            else:
                LOG.warning(
                    f"Case {case.internal_id} must be stored before creating a delivery report"
                )
        return stored_cases

    def get_cases_without_uploaded_delivery_report(self, pipeline: Pipeline) -> list[Case]:
        """Returns a list of cases that need a delivery report to be uploaded."""
        analyses: Query = self.status_db.analyses_to_upload_delivery_reports(pipeline=pipeline)[
            :MAX_ITEMS_TO_RETRIEVE
        ]
        return [analysis_obj.case for analysis_obj in analyses]

    def update_delivery_report_date(self, case: Case, analysis_date: datetime) -> None:
        """Updates the date when delivery report was created."""
        analysis: Analysis = self.status_db.get_analysis_by_case_entry_id_and_started_at(
            case_entry_id=case.id, started_at_date=analysis_date
        )
        analysis.delivery_report_created_at = datetime.now()
        self.status_db.session.commit()

    def get_report_data(self, case_id: str, analysis_date: datetime) -> ReportModel:
        """Fetches all the data needed to generate a delivery report."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        analysis: Analysis = self.status_db.get_analysis_by_case_entry_id_and_started_at(
            case_entry_id=case.id, started_at_date=analysis_date
        )
        analysis_metadata: AnalysisModel = self.analysis_api.get_latest_metadata(case.internal_id)
        case_model: CaseModel = self.get_case_data(case, analysis, analysis_metadata)
        return ReportModel(
            customer=self.get_customer_data(case=case),
            version=self.get_report_version(analysis=analysis),
            date=datetime.today(),
            case=case_model,
            accredited=self.is_report_accredited(
                samples=case_model.samples, analysis_metadata=analysis_metadata
            ),
        )

    def validate_report_fields(
        self, case_id: str, report_data: ReportModel, force_report
    ) -> ReportModel:
        """Verifies that the required report fields are not empty."""
        required_fields: dict = self.get_required_fields(case=report_data.case)
        empty_report_fields: dict = get_empty_report_data(report_data=report_data)
        missing_report_fields: dict = get_missing_report_data(
            empty_fields=empty_report_fields, required_fields=required_fields
        )
        if missing_report_fields and not force_report:
            LOG.error(
                f"Could not generate report data for {case_id}. "
                f"Missing data: \n{WriteStream.write_stream_from_content(content=missing_report_fields, file_format=FileFormat.YAML)}"
            )
            raise DeliveryReportError
        if empty_report_fields:
            LOG.warning(
                f"Empty report fields: \n{WriteStream.write_stream_from_content(content=empty_report_fields, file_format=FileFormat.YAML)}"
            )
        return report_data

    @staticmethod
    def get_customer_data(case: Case) -> CustomerModel:
        """Returns customer validated attributes retrieved from status DB."""
        return CustomerModel(
            name=case.customer.name,
            id=case.customer.internal_id,
            invoice_address=case.customer.invoice_address,
            scout_access=case.customer.scout_access,
        )

    @staticmethod
    def get_report_version(analysis: Analysis) -> int:
        """
        Returns the version of the given analysis. The version of the first analysis is 1 and subsequent reruns
        increase it by 1.
        """
        version = None
        if analysis:
            version = len(analysis.case.analyses) - analysis.case.analyses.index(analysis)
        return version

    def get_case_data(
        self,
        case: Case,
        analysis: Analysis,
        analysis_metadata: AnalysisModel,
    ) -> CaseModel:
        """Returns case associated validated attributes."""
        samples: list[SampleModel] = self.get_samples_data(
            case=case, analysis_metadata=analysis_metadata
        )
        unique_applications: list[ApplicationModel] = self.get_unique_applications(samples=samples)
        return CaseModel(
            name=case.name,
            id=case.internal_id,
            data_analysis=self.get_case_analysis_data(
                case=case, analysis=analysis, analysis_metadata=analysis_metadata
            ),
            samples=samples,
            applications=unique_applications,
        )

    def get_samples_data(self, case: Case, analysis_metadata: AnalysisModel) -> list[SampleModel]:
        """Extracts all the samples associated to a specific case and their attributes."""
        samples = list()
        case_samples: list[CaseSample] = self.status_db.get_case_samples_by_case_id(
            case_internal_id=case.internal_id
        )
        for case_sample in case_samples:
            sample: Sample = case_sample.sample
            lims_sample: dict | None = self.get_lims_sample(sample_id=sample.internal_id)
            samples.append(
                SampleModel(
                    name=sample.name,
                    id=sample.internal_id,
                    ticket=sample.original_ticket,
                    gender=sample.sex,
                    source=lims_sample.get("source") if lims_sample else None,
                    tumour=sample.is_tumour,
                    application=self.get_sample_application_data(lims_sample=lims_sample),
                    methods=self.get_sample_methods_data(sample_id=sample.internal_id),
                    status=case_sample.status,
                    metadata=self.get_sample_metadata(
                        case=case, sample=sample, analysis_metadata=analysis_metadata
                    ),
                    timestamps=self.get_sample_timestamp_data(sample=sample),
                )
            )
        return samples

    def get_lims_sample(self, sample_id: str) -> dict | None:
        """Fetches sample data from LIMS. Returns an empty dictionary if the request was unsuccessful."""
        lims_sample = dict()
        try:
            lims_sample: dict = self.lims_api.sample(sample_id)
        except requests.exceptions.HTTPError as ex:
            LOG.info(f"Could not fetch sample {sample_id} from LIMS: {ex}")
        return lims_sample

    def get_pipeline_accreditation_limitation(self, application_tag: str) -> str | None:
        """Return pipeline specific limitations given an application tag."""
        application_limitation: ApplicationLimitations = (
            self.status_db.get_application_limitation_by_tag_and_pipeline(
                tag=application_tag, pipeline=self.analysis_api.pipeline
            )
        )
        return application_limitation.limitations if application_limitation else None

    def get_sample_application_data(self, lims_sample: dict) -> ApplicationModel:
        """Retrieves the analysis application attributes."""
        application: Application = self.status_db.get_application_by_tag(
            tag=lims_sample.get("application")
        )
        return (
            ApplicationModel(
                tag=application.tag,
                version=lims_sample.get("application_version"),
                prep_category=application.prep_category,
                description=application.description,
                details=application.details,
                limitations=application.limitations,
                pipeline_limitations=self.get_pipeline_accreditation_limitation(application.tag),
                accredited=application.is_accredited,
                external=application.is_external,
            )
            if application
            else ApplicationModel()
        )

    @staticmethod
    def get_unique_applications(samples: list[SampleModel]) -> list[ApplicationModel]:
        """Returns the unique case associated applications."""
        applications = list()
        for sample in samples:
            if sample.application not in applications:
                applications.append(sample.application)
        return applications

    def get_sample_methods_data(self, sample_id: str) -> MethodsModel:
        """Fetches sample library preparation and sequencing methods from LIMS."""
        library_prep = None
        sequencing = None
        try:
            library_prep = self.lims_api.get_prep_method(lims_id=sample_id)
            sequencing = self.lims_api.get_sequencing_method(lims_id=sample_id)
        except requests.exceptions.HTTPError as ex:
            LOG.info(f"Could not fetch sample ({sample_id}) methods from LIMS: {ex}")

        return MethodsModel(library_prep=library_prep, sequencing=sequencing)

    def get_case_analysis_data(
        self,
        case: Case,
        analysis: Analysis,
        analysis_metadata: AnalysisModel,
    ) -> DataAnalysisModel:
        """Return pipeline attributes used for data analysis."""
        return DataAnalysisModel(
            customer_pipeline=case.data_analysis,
            data_delivery=case.data_delivery,
            pipeline=analysis.pipeline,
            pipeline_version=analysis.pipeline_version,
            type=self.get_data_analysis_type(case=case),
            genome_build=self.get_genome_build(analysis_metadata=analysis_metadata),
            variant_callers=self.get_variant_callers(_analysis_metadata=analysis_metadata),
            panels=case.panels,
            scout_files=self.get_scout_uploaded_files(case=case),
        )

    def get_scout_uploaded_files(self, case: Case) -> ScoutReportFiles:
        """Return files that will be uploaded to Scout."""
        return ScoutReportFiles(
            snv_vcf=self.get_scout_uploaded_file_from_hk(
                case_id=case.internal_id, scout_tag="snv_vcf"
            ),
            sv_vcf=self.get_scout_uploaded_file_from_hk(
                case_id=case.internal_id, scout_tag="sv_vcf"
            ),
            vcf_str=self.get_scout_uploaded_file_from_hk(
                case_id=case.internal_id, scout_tag="vcf_str"
            ),
            smn_tsv=self.get_scout_uploaded_file_from_hk(
                case_id=case.internal_id, scout_tag="smn_tsv"
            ),
            vcf_fusion=self.get_scout_uploaded_file_from_hk(
                case_id=case.internal_id, scout_tag="vcf_fusion"
            ),
        )

    @staticmethod
    def get_sample_timestamp_data(sample: Sample) -> TimestampModel:
        """Return sample processing dates."""
        return TimestampModel(
            ordered_at=sample.ordered_at,
            received_at=sample.received_at,
            prepared_at=sample.prepared_at,
            reads_updated_at=sample.last_sequenced_at,
        )

    def get_sample_metadata(
        self,
        case: Case,
        sample: Sample,
        analysis_metadata: AnalysisModel,
    ) -> SampleMetadataModel:
        """Return sample metadata to include in the report."""
        raise NotImplementedError

    def get_data_analysis_type(self, case: Case) -> str | None:
        """Return data analysis type carried out."""
        case_sample: Sample = self.status_db.get_case_samples_by_case_id(
            case_internal_id=case.internal_id
        )[0].sample
        lims_sample: dict | None = self.get_lims_sample(sample_id=case_sample.internal_id)
        application: Application = self.status_db.get_application_by_tag(
            tag=lims_sample.get("application")
        )
        return application.analysis_type if application else None

    def get_genome_build(self, analysis_metadata: AnalysisModel) -> str:
        """Return build version of the genome reference of a specific case."""
        raise NotImplementedError

    def get_variant_callers(self, _analysis_metadata: AnalysisModel) -> list:
        """Return list of variant-calling filters used during analysis."""
        return []

    def is_report_accredited(
        self, samples: list[SampleModel], analysis_metadata: AnalysisModel
    ) -> bool:
        """Check if the report is accredited."""
        raise NotImplementedError

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return dictionary with the delivery report required fields."""
        raise NotImplementedError

    def get_template_name(self) -> str:
        """Return pipeline specific template name."""
        raise NotImplementedError

    @staticmethod
    def get_application_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Return sample required fields."""
        required_sample_fields = dict()
        for application in case.applications:
            required_sample_fields.update({application.tag: required_fields})
        return required_sample_fields

    @staticmethod
    def get_sample_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Return sample required fields."""
        required_sample_fields = dict()
        for sample in case.samples:
            required_sample_fields.update({sample.id: required_fields})
        return required_sample_fields

    @staticmethod
    def get_timestamp_required_fields(case: CaseModel, required_fields: list) -> dict:
        """Return sample timestamps required fields."""
        for sample in case.samples:
            if sample.application.external:
                required_fields.remove("received_at")
                break
        return ReportAPI.get_sample_required_fields(case=case, required_fields=required_fields)

    def get_hk_scout_file_tags(self, scout_tag: str) -> list | None:
        """Return pipeline specific uploaded to Scout Housekeeper file tags given a Scout key."""
        tags = self.get_upload_case_tags().get(scout_tag)
        return list(tags) if tags else None

    def get_upload_case_tags(self):
        """Return pipeline specific upload case tags."""
        raise NotImplementedError



import logging
from pathlib import Path

from click.testing import CliRunner
from psutil import Process

from cg.cli.backup import backup_flow_cells, encrypt_flow_cells, fetch_flow_cell
from cg.constants import EXIT_SUCCESS, FileExtensions, FlowCellStatus
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from tests.store_helpers import StoreHelpers


def test_backup_flow_cells(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    flow_cell_name: str,
    flow_cell_full_name: str,
    helpers: StoreHelpers,
):
    """Test backing up flow cell in dry run mode."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a flow cells directory

    # Given a flow cell with no back-up
    helpers.add_flow_cell(
        store=cg_context.status_db, flow_cell_name=flow_cell_name, has_backup=False
    )

    # GIVEN an encrypted flow cell
    flow_cells_encrypt_dir = Path(cg_context.encryption.encryption_dir, flow_cell_full_name)
    flow_cells_encrypt_dir.mkdir(parents=True, exist_ok=True)
    Path(flow_cells_encrypt_dir, flow_cell_name).with_suffix(FileExtensions.COMPLETE).touch()

    # WHEN backing up flow cells in dry run mode
    result = cli_runner.invoke(backup_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS


def test_backup_flow_cells_when_dsmc_is_running(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    flow_cell_name: str,
    flow_cell_full_name: str,
    mocker,
):
    """Test backing-up flow cell in dry run mode when Dsmc processing has started."""
    caplog.set_level(logging.ERROR)

    # GIVEN a flow cells directory

    # GIVEN an ongoing Dsmc process
    mocker.patch.object(Process, "name", return_value="dsmc")

    # WHEN backing up flow cells in dry run mode
    result = cli_runner.invoke(backup_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate too many Dsmc processes are already running
    assert "Too many Dsmc processes are already running" in caplog.text


def test_backup_flow_cells_when_flow_cell_already_has_backup(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    flow_cell_name: str,
    flow_cell_full_name: str,
    helpers: StoreHelpers,
):
    """Test backing-up flow cell in dry run mode when already backed-up."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a flow cells directory

    # GIVEN a flow cell with a back-up
    helpers.add_flow_cell(
        store=cg_context.status_db, flow_cell_name=flow_cell_name, has_backup=True
    )

    # WHEN backing up flow cells in dry run mode
    result = cli_runner.invoke(backup_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell has already benn backed upped
    assert f"Flow cell: {flow_cell_name} is already backed-up" in caplog.text


def test_backup_flow_cells_when_encryption_is_not_completed(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    flow_cell_name: str,
    flow_cell_full_name: str,
):
    """Test backing-up flow cell in dry run mode when encryption is not complete."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a flow cells directory

    # WHEN backing up flow cells in dry run mode
    result = cli_runner.invoke(backup_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell encryption is not completed
    assert f"Flow cell: {flow_cell_name} encryption process is not complete" in caplog.text


def test_encrypt_flow_cells(
    cli_runner: CliRunner, cg_context: CGConfig, caplog, sbatch_job_number: str
):
    """Test encrypt flow cell in dry run mode."""
    caplog.set_level(logging.INFO)

    # GIVEN a flow cells directory

    # WHEN encrypting flow cells in dry run mode
    result = cli_runner.invoke(encrypt_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell encryption is submitted
    assert f"Flow cell encryption running as job {sbatch_job_number}" in caplog.text


def test_encrypt_flow_cells_when_already_backed_up(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    flow_cell_name: str,
    helpers: StoreHelpers,
):
    """Test encrypt flow cell in dry run mode when there is already a back-up."""
    caplog.set_level(logging.DEBUG)

    # Given a flow cell with a back-up
    helpers.add_flow_cell(
        store=cg_context.status_db, flow_cell_name=flow_cell_name, has_backup=True
    )

    # GIVEN a flow cells directory

    # WHEN encrypting flow cells in dry run mode
    result = cli_runner.invoke(encrypt_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell is already backed-up
    assert f"Flow cell: {flow_cell_name} is already backed-up" in caplog.text


def test_encrypt_flow_cells_when_sequencing_not_done(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    mocker,
    flow_cell_name: str,
):
    """Test encrypt flow cell in dry run mode when sequencing is not done."""
    caplog.set_level(logging.DEBUG)

    # GIVEN flow cells that are being sequenced
    mocker.patch.object(FlowCellDirectoryData, "is_flow_cell_ready")
    FlowCellDirectoryData.is_flow_cell_ready.return_value = False

    # GIVEN a flow cells directory

    # WHEN encrypting flow cells in dry run mode
    result = cli_runner.invoke(encrypt_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell is not ready
    assert f"Flow cell: {flow_cell_name} is not ready" in caplog.text


def test_encrypt_flow_cell_when_encryption_already_started(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    pdc_archiving_dir: Path,
    flow_cell_name: str,
    flow_cell_full_name: str,
    mocker,
):
    """Test encrypt flow cell in dry run mode when pending file exists"""
    caplog.set_level(logging.DEBUG)

    # GIVEN flow cells that are ready
    mocker.patch.object(FlowCellDirectoryData, "is_flow_cell_ready")
    FlowCellDirectoryData.is_flow_cell_ready.return_value = True

    # GIVEN a pending flag file
    flow_cells_encrypt_dir = Path(cg_context.encryption.encryption_dir, flow_cell_full_name)
    flow_cells_encrypt_dir.mkdir(parents=True, exist_ok=True)
    Path(flow_cells_encrypt_dir, flow_cell_name).with_suffix(FileExtensions.PENDING).touch()

    # GIVEN a flow cells directory

    # WHEN encrypting flow cells in dry run mode
    result = cli_runner.invoke(encrypt_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell encryption already started
    assert f"Encryption already started for flow cell: {flow_cell_name}" in caplog.text


def test_encrypt_flow_cell_when_encryption_already_completed(
    cli_runner: CliRunner,
    cg_context: CGConfig,
    caplog,
    pdc_archiving_dir: Path,
    flow_cell_name: str,
    flow_cell_full_name: str,
    mocker,
):
    """Test encrypt flow cell in dry run mode when completed file exists"""
    caplog.set_level(logging.DEBUG)

    # GIVEN flow cells that are ready
    mocker.patch.object(FlowCellDirectoryData, "is_flow_cell_ready")
    FlowCellDirectoryData.is_flow_cell_ready.return_value = True

    # GIVEN a complete flag file
    flow_cells_encrypt_dir = Path(cg_context.encryption.encryption_dir, flow_cell_full_name)
    flow_cells_encrypt_dir.mkdir(parents=True, exist_ok=True)
    Path(flow_cells_encrypt_dir, flow_cell_name).with_suffix(FileExtensions.COMPLETE).touch()

    # GIVEN a flow cells directory

    # WHEN encrypting flow cells in dry run mode
    result = cli_runner.invoke(encrypt_flow_cells, ["--dry-run"], obj=cg_context)

    # THEN exits without any errors
    assert result.exit_code == EXIT_SUCCESS

    # THEN communicate flow cell encryption already completed
    assert f"Encryption already completed for flow cell: {flow_cell_name}" in caplog.text


def test_run_fetch_flow_cell_dry_run_no_flow_cell_specified(
    cli_runner: CliRunner, backup_context: CGConfig, caplog
):
    """Test fetching flow cell when no flow cells with correct status."""
    caplog.set_level(logging.INFO)

    # GIVEN a context with a backup_api
    assert "backup_api" in backup_context.meta_apis

    # GIVEN that there are no flow cells set to "requested" in status_db
    assert not backup_context.status_db.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.REQUESTED]
    )

    # WHEN running the fetch flow cell command without specifying any flow cell in dry run mode
    result = cli_runner.invoke(fetch_flow_cell, ["--dry-run"], obj=backup_context)

    # THEN assert that it exits without any problems
    assert result.exit_code == EXIT_SUCCESS

    # THEN assert that it is communicated that no flow cells are requested
    assert "No flow cells requested" in caplog.text


def test_run_fetch_flow_cell_dry_run_retrieval_time(
    cli_runner: CliRunner, backup_context: CGConfig, caplog, mocker
):
    """Test fetching flow cell retrieval time."""
    caplog.set_level(logging.INFO)

    # GIVEN a context with a backup_api
    assert "backup_api" in backup_context.meta_apis

    # GIVEN that there are no flow cells set to "requested" in status_db
    assert not backup_context.status_db.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.REQUESTED]
    )

    # GIVEN that the backup api returns a retrieval time
    expected_time = 60
    mocker.patch("cg.meta.backup.backup.BackupAPI.fetch_flow_cell", return_value=expected_time)

    # WHEN running the fetch flow cell command without specifying any flow cell in dry run mode
    result = cli_runner.invoke(fetch_flow_cell, ["--dry-run"], obj=backup_context)

    # THEN assert that it exits without any problems
    assert result.exit_code == EXIT_SUCCESS

    # THEN assert that it is communicated that a retrieval time was found
    assert "Retrieval time" in caplog.text


def test_run_fetch_flow_cell_non_existing_flow_cell(
    cli_runner: CliRunner, backup_context: CGConfig, caplog
):
    # GIVEN a context with a backup api
    # GIVEN a non-existing flow cell id
    flow_cell_id = "hello"
    assert backup_context.status_db.get_flow_cell_by_name(flow_cell_id) is None

    # WHEN running the command with the non-existing flow cell id
    result = cli_runner.invoke(
        fetch_flow_cell, ["--flow-cell-id", flow_cell_id], obj=backup_context
    )

    # THEN assert that it exits with a non-zero exit code
    assert result.exit_code != 0
    # THEN assert that it was communicated that the flow cell does not exist
    assert f"{flow_cell_id}: not found" in caplog.text



"""Constants related to Oxford Nanopore sequencing."""
from enum import StrEnum


class NanoporeDirsAndFiles(StrEnum):
    """Holds file and directory names for nanopore sequencing output."""

    DATA_DIRECTORY: str = "nanopore"



"""Observations API."""

import logging
from datetime import datetime
from pathlib import Path

from housekeeper.store.models import Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.loqus import LoqusdbAPI
from cg.constants.observations import (
    LoqusdbBalsamicCustomers,
    LoqusdbInstance,
    LoqusdbMipCustomers,
)
from cg.exc import LoqusdbUploadCaseError
from cg.models.cg_config import CGConfig, CommonAppConfig
from cg.models.observations.input_files import (
    BalsamicObservationsInputFiles,
    MipDNAObservationsInputFiles,
)
from cg.store import Store
from cg.store.models import Analysis, Case, Customer

LOG = logging.getLogger(__name__)


class ObservationsAPI:
    """API to manage Loqusdb observations."""

    def __init__(self, config: CGConfig):
        self.store: Store = config.status_db
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.loqusdb_config: CommonAppConfig = config.loqusdb
        self.loqusdb_wes_config: CommonAppConfig = config.loqusdb_wes
        self.loqusdb_somatic_config: CommonAppConfig = config.loqusdb_somatic
        self.loqusdb_tumor_config: CommonAppConfig = config.loqusdb_tumor

    def upload(self, case: Case) -> None:
        """Upload observations to Loqusdb."""
        self.check_customer_loqusdb_permissions(case.customer)
        input_files: MipDNAObservationsInputFiles | BalsamicObservationsInputFiles = (
            self.get_observations_input_files(case)
        )
        self.load_observations(case=case, input_files=input_files)

    def get_observations_input_files(
        self, case: Case
    ) -> MipDNAObservationsInputFiles | BalsamicObservationsInputFiles:
        """Fetch input files from a case to upload to Loqusdb."""
        analysis: Analysis = case.analyses[0]
        analysis_date: datetime = analysis.started_at or analysis.completed_at
        hk_version: Version = self.housekeeper_api.version(analysis.case.internal_id, analysis_date)
        return self.extract_observations_files_from_hk(hk_version)

    def get_loqusdb_api(self, loqusdb_instance: LoqusdbInstance) -> LoqusdbAPI:
        """Returns a Loqusdb API for the given Loqusdb instance."""
        loqusdb_apis = {
            LoqusdbInstance.WGS: LoqusdbAPI(
                binary_path=self.loqusdb_config.binary_path,
                config_path=self.loqusdb_config.config_path,
            ),
            LoqusdbInstance.WES: LoqusdbAPI(
                binary_path=self.loqusdb_wes_config.binary_path,
                config_path=self.loqusdb_wes_config.config_path,
            ),
            LoqusdbInstance.SOMATIC: LoqusdbAPI(
                binary_path=self.loqusdb_somatic_config.binary_path,
                config_path=self.loqusdb_somatic_config.config_path,
            ),
            LoqusdbInstance.TUMOR: LoqusdbAPI(
                binary_path=self.loqusdb_tumor_config.binary_path,
                config_path=self.loqusdb_tumor_config.config_path,
            ),
        }
        return loqusdb_apis[loqusdb_instance]

    @staticmethod
    def is_duplicate(
        case: Case,
        loqusdb_api: LoqusdbAPI,
        profile_vcf_path: Path | None,
        profile_threshold: float | None,
    ) -> bool:
        """Check if a case has already been uploaded to Loqusdb."""
        loqusdb_case: dict = loqusdb_api.get_case(case_id=case.internal_id)
        duplicate = (
            loqusdb_api.get_duplicate(
                profile_vcf_path=profile_vcf_path, profile_threshold=profile_threshold
            )
            if profile_vcf_path and profile_threshold
            else None
        )
        return bool(loqusdb_case or duplicate or case.loqusdb_uploaded_samples)

    def update_statusdb_loqusdb_id(self, samples: list[Case], loqusdb_id: str | None) -> None:
        """Update Loqusdb ID field in StatusDB for each of the provided samples."""
        for sample in samples:
            sample.loqusdb_id = loqusdb_id
        self.store.session.commit()

    def check_customer_loqusdb_permissions(self, customer: Customer) -> None:
        """Verifies that the customer is whitelisted for Loqusdb uploads."""
        if customer.internal_id not in [cust_id for cust_id in self.get_loqusdb_customers()]:
            LOG.error(
                f"Customer {customer.internal_id} is not whitelisted for Loqusdb uploads. Cancelling upload."
            )
            raise LoqusdbUploadCaseError
        LOG.info(f"Valid customer {customer.internal_id} for Loqusdb uploads")

    def get_loqusdb_customers(self) -> LoqusdbMipCustomers | LoqusdbBalsamicCustomers:
        """Returns the customers that are entitled to Loqusdb uploads."""
        raise NotImplementedError

    def load_observations(
        self,
        case: Case,
        input_files: MipDNAObservationsInputFiles | BalsamicObservationsInputFiles,
    ) -> None:
        """Load observation counts to Loqusdb."""
        raise NotImplementedError

    def extract_observations_files_from_hk(
        self, hk_version: Version
    ) -> MipDNAObservationsInputFiles | BalsamicObservationsInputFiles:
        """Extract observations files given a housekeeper version."""
        raise NotImplementedError

    def delete_case(self, case: Case) -> None:
        """Delete case observations from Loqusdb."""
        raise NotImplementedError



from pathlib import Path

import pytest

from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreatorBcl2Fastq,
    SampleSheetCreatorBCLConvert,
)
from cg.constants.demultiplexing import SampleSheetBcl2FastqSections
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData


@pytest.fixture
def bcl_convert_samples_similar_index1() -> list[FlowCellSampleBCLConvert]:
    """Return a list of three FlowCellSampleBCLConvert with updated indexes."""
    sample_1 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC123", index="CAGAAGAT", index2="GCGCAAGC"
    )
    sample_2 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC456", index="CAGAAGAG", index2="CAATGTAT"
    )
    sample_3 = FlowCellSampleBCLConvert(
        lane=2, sample_id="ACC789", index="AAGCGATA", index2="AACCGCAA"
    )
    return [sample_1, sample_2, sample_3]


@pytest.fixture
def bcl_convert_samples_similar_index2() -> list[FlowCellSampleBCLConvert]:
    """Return a list of three FlowCellSampleBCLConvert with updated indexes."""
    sample_1 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC123", index="GCGCAAGC", index2="CAATGTAC"
    )
    sample_2 = FlowCellSampleBCLConvert(
        lane=1, sample_id="ACC456", index="CAATGTAT", index2="CAATGTAT"
    )
    sample_3 = FlowCellSampleBCLConvert(
        lane=2, sample_id="ACC789", index="AAGCGATA", index2="AACCGCAA"
    )
    return [sample_1, sample_2, sample_3]


@pytest.fixture
def bcl2fastq_sample_sheet_creator(
    novaseq_flow_cell_demultiplexed_with_bcl2fastq: FlowCellDirectoryData,
    lims_novaseq_6000_bcl2fastq_samples: list[FlowCellSampleBcl2Fastq],
) -> SampleSheetCreatorBcl2Fastq:
    """Returns a sample sheet creator for version 1 sample sheets with bcl2fastq format."""
    return SampleSheetCreatorBcl2Fastq(
        flow_cell=novaseq_flow_cell_demultiplexed_with_bcl2fastq,
        lims_samples=lims_novaseq_6000_bcl2fastq_samples,
    )


# Sample sheet validation


@pytest.fixture
def sample_sheet_line_sample_1() -> list[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A68",
        "hg19",
        "ATTCCACACT",
        "TGGTCTTGTT",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture
def sample_sheet_line_sample_2() -> list[str]:
    """Return the line in the sample sheet corresponding to a sample."""
    return [
        "HWHMWDMXX",
        "1",
        "ACC7628A1",
        "hg19",
        "AGTTAGCTGG",
        "GATGAGAATG",
        "814206",
        "N",
        "R1",
        "script",
        "814206",
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_data_header() -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet data header without samples."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBcl2FastqSections.Data.LANE.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_REFERENCE.value,
            SampleSheetBcl2FastqSections.Data.INDEX_1.value,
            SampleSheetBcl2FastqSections.Data.INDEX_2.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_NAME.value,
            SampleSheetBcl2FastqSections.Data.CONTROL.value,
            SampleSheetBcl2FastqSections.Data.RECIPE.value,
            SampleSheetBcl2FastqSections.Data.OPERATOR.value,
            SampleSheetBcl2FastqSections.Data.SAMPLE_PROJECT_BCL2FASTQ.value,
        ],
    ]


@pytest.fixture
def sample_sheet_samples_no_header(
    sample_sheet_line_sample_1: list[str], sample_sheet_line_sample_2: list[str]
) -> list[list[str]]:
    """Return the content of a sample sheet with samples but without a sample header."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def valid_sample_sheet_bcl2fastq(
    sample_sheet_bcl2fastq_data_header: list[list[str]],
    sample_sheet_line_sample_1: list[str],
    sample_sheet_line_sample_2: list[str],
) -> list[list[str]]:
    """Return the content of a valid Bcl2fastq sample sheet."""
    return sample_sheet_bcl2fastq_data_header + [
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def sample_sheet_bcl2fastq_duplicate_same_lane(
    valid_sample_sheet_bcl2fastq: list[list[str]], sample_sheet_line_sample_2: list[str]
) -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in the same lane."""
    valid_sample_sheet_bcl2fastq.append(sample_sheet_line_sample_2)
    return valid_sample_sheet_bcl2fastq


@pytest.fixture
def sample_sheet_bcl2fastq_duplicate_different_lane(
    valid_sample_sheet_bcl2fastq: list[list[str]],
) -> list[list[str]]:
    """Return the content of a Bcl2fastq sample sheet with a duplicated sample in a different lane."""
    valid_sample_sheet_bcl2fastq.append(
        [
            "HWHMWDMXX",
            "2",
            "ACC7628A1",
            "hg19",
            "AGTTAGCTGG",
            "GATGAGAATG",
            "814206",
            "N",
            "R1",
            "script",
            "814206",
        ]
    )
    return valid_sample_sheet_bcl2fastq


@pytest.fixture
def valid_sample_sheet_dragen(
    sample_sheet_line_sample_1: list[str], sample_sheet_line_sample_2: list[str]
) -> list[list[str]]:
    """Return the content of a valid Dragen sample sheet."""
    return [
        [SampleSheetBcl2FastqSections.Data.HEADER],
        [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            "Lane",
            "Sample_ID",
            "SampleRef",
            "index",
            "index2",
            "SampleName",
            "Control",
            "Recipe",
            "Operator",
            "Sample_Project",
        ],
        sample_sheet_line_sample_1,
        sample_sheet_line_sample_2,
    ]


@pytest.fixture
def sample_sheet_dragen_duplicate_same_lane(
    valid_sample_sheet_dragen: list[list[str]], sample_sheet_line_sample_2: list[str]
) -> list[list[str]]:
    """Return the content of a Dragen sample sheet with a duplicated sample in the same lane."""
    valid_sample_sheet_dragen.append(sample_sheet_line_sample_2)
    return valid_sample_sheet_dragen


@pytest.fixture
def sample_sheet_dragen_duplicate_different_lane(
    valid_sample_sheet_dragen: list[list[str]],
) -> list[list[str]]:
    """Return the content of a Dragen sample sheet with a duplicated sample in a different lane."""
    valid_sample_sheet_dragen.append(
        [
            "HWHMWDMXX",
            "2",
            "ACC7628A1",
            "hg19",
            "AGTTAGCTGG",
            "GATGAGAATG",
            "814206",
            "N",
            "R1",
            "script",
            "814206",
        ]
    )
    return valid_sample_sheet_dragen


@pytest.fixture
def novaseq6000_flow_cell_sample_1() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=1,
        SampleID="ACC7628A68",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture
def novaseq6000_flow_cell_sample_2() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        SampleRef="hg19",
        index="ATTCCACACT",
        index2="TGGTCTTGTT",
        SampleName="814206",
        Control="N",
        Recipe="R1",
        Operator="script",
        Project="814206",
    )


@pytest.fixture
def novaseq6000_flow_cell_sample_no_dual_index() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT",
        SampleName="814206",
        Project="814206",
    )


@pytest.fixture
def novaseq6000_flow_cell_sample_before_adapt_indexes() -> FlowCellSampleBcl2Fastq:
    """Return a NovaSeq sample without dual indexes."""
    return FlowCellSampleBcl2Fastq(
        FCID="HWHMWDMXX",
        Lane=2,
        SampleID="ACC7628A1",
        index="ATTCCACACT-TGGTCTTGTT",
        SampleName="814206",
        Project="814206",
    )


@pytest.fixture
def index1_sequence_from_lims() -> str:
    """Return an index 1 sequence."""
    return "GTCTACAC"


@pytest.fixture
def index2_sequence_from_lims() -> str:
    """Return an index 2 sequence."""
    return "GCCAAGGT"


@pytest.fixture
def raw_index_sequence(index1_sequence_from_lims: str, index2_sequence_from_lims: str) -> str:
    """Return a raw index."""
    return f"{index1_sequence_from_lims}-{index2_sequence_from_lims}"


@pytest.fixture
def bcl_convert_flow_cell_sample(raw_index_sequence: str) -> FlowCellSampleBCLConvert:
    """Return a BCL Convert sample."""
    return FlowCellSampleBCLConvert(lane=1, index=raw_index_sequence, sample_id="ACC123")


@pytest.fixture
def bcl_convert_sample_sheet_path(demultiplexed_runs: Path):
    return Path(
        demultiplexed_runs,
        "230504_A00689_0804_BHY7FFDRX2",
        "SampleSheet.csv",
    )


@pytest.fixture
def bcl2fastq_sample_sheet_path(demultiplexed_runs: Path):
    return Path(
        demultiplexed_runs,
        "170407_ST-E00198_0209_BHHKVCALXX",
        "SampleSheet.csv",
    )
